@@ -40,7 +40,6 @@ global ws_label
 global http_server
 scheduler = None
 current_slideshow = None
-CACHEDIR = r'''C:\Windows\Temp\ppt-cache'''
 interface_root = None
 CONFIG_FILE = r'''..\ppt-control.ini'''
 LOGFILE = r'''..\ppt-control.log'''
@@ -81,7 +80,7 @@ class Handler(server.SimpleHTTPRequestHandler):
             black = 0
             if current_slideshow:
                 try:
-                    path = CACHEDIR + "\\" + current_slideshow.name()
+                    path = config.prefs["Main"]["cache"] + "\\" + current_slideshow.name()
                 except Exception as e:
                     path = os.path.join(os.path.dirname(os.path.realpath(__file__)), r'''\static\black.jpg''') + '/'
                     logger.warning("Failed to get current slideshow name: ", e)
@@ -103,7 +102,7 @@ class Handler(server.SimpleHTTPRequestHandler):
 
 def run_http():
     global http_server
-    http_server = server.HTTPServer(("", 80), Handler)
+    http_server = server.HTTPServer((config.prefs["HTTP"]["interface"], config.prefs.getint("HTTP", "port")), Handler)
     http_server.serve_forever()
 
 STATE_DEFAULT = {"connected": 0, "current": 0, "total": 0, "visible": 0, "name": ""}
@@ -115,11 +114,7 @@ def state_event():
     return json.dumps({"type": "state", **STATE})
 
 
-def users_event():
-    return json.dumps({"type": "users", "count": len(USERS)})
-
-
-async def notify_state():
+def notify_state():
     global STATE
     if current_slideshow and STATE["connected"] == 1:
         try:
@@ -134,85 +129,85 @@ async def notify_state():
         STATE = copy(STATE_DEFAULT)
     if USERS:  # asyncio.wait doesn't accept an empty list
         message = state_event()
-        await asyncio.wait([user.send(message) for user in USERS])
+        loop.call_soon_threadsafe(ws_queue.put_nowait, state_event())
 
 
-async def notify_users():
-    if USERS:  # asyncio.wait doesn't accept an empty list
-        message = users_event()
-        await asyncio.wait([user.send(message) for user in USERS])
 
+async def ws_handler(websocket, path):
+    logger.debug("Handling WebSocket connection")
+    recv_task = asyncio.ensure_future(ws_receive(websocket, path))
+    send_task = asyncio.ensure_future(ws_send(websocket, path))
+    done, pending = await asyncio.wait(
+        [recv_task, send_task],
+        return_when=asyncio.FIRST_COMPLETED,
+    )
+    for task in pending:
+        task.cancel()
 
-async def register(websocket):
-    USERS.add(websocket)
-    await notify_users()
-
-
-async def unregister(websocket):
-    USERS.remove(websocket)
-    await notify_users()
-
-
-async def ws_handle(websocket, path):
+async def ws_receive(websocket, path):
     logger.debug("Received websocket request")
-    global current_slideshow
-    # register(websocket) sends user_event() to websocket
-    await register(websocket)
+    USERS.add(websocket)
     try:
-        await websocket.send(state_event())
+        # Send initial state to clients on load
+        notify_state()
         async for message in websocket:
+            logger.debug("Received websocket message: " + str(message))
             data = json.loads(message)
             if data["action"] == "prev":
                 if current_slideshow:
                     current_slideshow.prev()
-                await notify_state()
+                notify_state()
             elif data["action"] == "next":
                 if current_slideshow:
                     current_slideshow.next()
-                await notify_state()
+                notify_state()
             elif data["action"] == "first":
                 if current_slideshow:
                     current_slideshow.first()
-                await notify_state()
+                notify_state()
             elif data["action"] == "last":
                 if current_slideshow:
                     current_slideshow.last()
-                await notify_state()
+                notify_state()
             elif data["action"] == "black":
                 if current_slideshow:
                     if current_slideshow.visible() == 3:
                         current_slideshow.normal()
                     else:
                         current_slideshow.black()
-                await notify_state()
+                notify_state()
             elif data["action"] == "white":
                 if current_slideshow:
                     if current_slideshow.visible() == 4:
                         current_slideshow.normal()
                     else:
                         current_slideshow.white()
-                await notify_state()
+                notify_state()
             elif data["action"] == "goto":
                 if current_slideshow:
                     current_slideshow.goto(int(data["value"]))
-                await notify_state()
-            elif data["action"] == "refresh":
-                await notify_state()
-                if current_slideshow:
-                    current_slideshow.export_current_next()
-                    current_slideshow.refresh()
+                notify_state()
             else:
-                logger.error("unsupported event: {}", data)
+                logger.error("Received unnsupported event: {}", data)
     finally:
-        await unregister(websocket)
+        USERS.remove(websocket)
+
+async def ws_send(websocket, path):
+    while True:
+        message = await ws_queue.get()
+        await asyncio.wait([user.send(message) for user in USERS])
+
 
 def run_ws():
     # https://stackoverflow.com/questions/21141217/how-to-launch-win32-applications-in-separate-threads-in-python/22619084#22619084
     # https://www.reddit.com/r/learnpython/comments/mwt4qi/pywintypescom_error_2147417842_the_application/
     pythoncom.CoInitializeEx(pythoncom.COINIT_MULTITHREADED)
     asyncio.set_event_loop(asyncio.new_event_loop())
-    #start_server = websockets.serve(ws_handle, "0.0.0.0", 5678, ping_interval=None)
-    start_server = websockets.serve(ws_handle, "0.0.0.0", 5678)
+    global ws_queue
+    ws_queue = asyncio.Queue()
+    global loop
+    loop = asyncio.get_event_loop()
+    start_server = websockets.serve(ws_handler, config.prefs["WebSocket"]["interface"], config.prefs.getint("WebSocket", "port"), ping_interval=None)
     asyncio.get_event_loop().run_until_complete(start_server)
     asyncio.get_event_loop().run_forever()
 
@@ -239,7 +234,7 @@ def start_ws():
     logger.info("Started websocket server")
 
 class Slideshow:
-    def __init__(self, instance):
+    def __init__(self, instance, blackwhite):
         self.instance = instance
         if self.instance is None:
             raise ValueError("PPT instance cannot be None")
@@ -252,13 +247,14 @@ class Slideshow:
             raise ValueError("PPT instance has no  active presentation")
         self.presentation = self.instance.ActivePresentation
 
+        self.blackwhite = blackwhite
+
         self.export_current_next()
 
     def unload(self):
         connect_ppt()
 
     def refresh(self):
-        logger.debug("Refreshing")
         try:
             if self.instance is None:
                 raise ValueError("PPT instance cannot be None")
@@ -340,7 +336,10 @@ class Slideshow:
     def black(self):
         try:
             self.refresh()
-            self.view.State = 3
+            if self.blackwhite == "both" and self.view.State == 4:
+                self.view.state = 1
+            else:
+                self.view.State = 3
             self.export_current_next()
         except (ValueError, pywintypes.com_error):
             self.unload()
@@ -348,7 +347,10 @@ class Slideshow:
     def white(self):
         try:
             self.refresh()
-            self.view.State = 4
+            if self.blackwhite == "both" and self.view.State == 3:
+                self.view.state = 1
+            else:
+                self.view.State = 4
             self.export_current_next()
         except (ValueError, pywintypes.com_error):
             self.unload()
@@ -375,20 +377,23 @@ class Slideshow:
         self.export(self.current_slide() + 2)
 
     def export(self, slide):
-        destination = CACHEDIR + "\\" + self.name() + "\\" + str(slide) + ".jpg"
+        destination = config.prefs["Main"]["cache"] + "\\" + self.name() + "\\" + str(slide) + ".jpg"
         os.makedirs(os.path.dirname(destination), exist_ok=True)
         if not os.path.exists(destination) or time.time() - os.path.getmtime(destination) > config.prefs.getint("Main", "cache_timeout"):
             if slide <= self.total_slides():
                 attempts = 0
                 while attempts < 3:
                     try:
-                        self.presentation.Slides(slide).Export(destination, "JPG")
+                        self.presentation.Slides(slide).Export(destination, config.prefs["Main"]["cache_format"])
                         break
                     except:
                         pass
                     attempts += 1
             elif slide == self.total_slides() + 1:
-                shutil.copyfileobj(open(os.path.dirname(os.path.realpath(__file__)), r'''\static\black.jpg''', 'rb'), open(destination, 'wb'))
+                try:
+                    shutil.copyfileobj(open(os.path.dirname(os.path.realpath(__file__)), r'''\static\black.jpg''', 'rb'), open(destination, 'wb'))
+                except Exception as e:
+                    logger.warning("Failed to copy black slide: " + str(e))
             else:
                 pass
 
@@ -407,10 +412,10 @@ def get_current_slideshow():
 
 def refresh_interval():
     while getattr(refresh_daemon, "do_run", True):
-        logger.debug("Triggering server-side refresh")
         current_slideshow.refresh()
+        notify_state()
         refresh_status()
-        time.sleep(1)
+        time.sleep(0.5)
 
 def refresh_status():
     if status_label is not None:
@@ -431,7 +436,7 @@ def connect_ppt():
         try:
             instance = get_ppt_instance()
             global current_slideshow
-            current_slideshow = Slideshow(instance)
+            current_slideshow = Slideshow(instance, config.prefs["Main"]["blackwhite"])
             STATE["connected"] = 1
             STATE["current"] = current_slideshow.current_slide()
             STATE["total"] = current_slideshow.total_slides()
@@ -564,10 +569,10 @@ def start_interface():
     console_handler.setLevel(log_level)
     logger.addHandler(console_handler)
 
-    logging.getLogger("asyncio").setLevel(logging.ERROR)
-    logging.getLogger("asyncio.coroutines").setLevel(logging.ERROR)
+    #logging.getLogger("asyncio").setLevel(logging.ERROR)
+    #logging.getLogger("asyncio.coroutines").setLevel(logging.ERROR)
     logging.getLogger("websockets.server").setLevel(logging.ERROR)
-    logging.getLogger("websockets.protocol").setLevel(logging.ERROR)
+    #logging.getLogger("websockets.protocol").setLevel(logging.ERROR)
 
 
     logger.debug("Finished setting up config and logging")
